from xvc import Xvc
import pytest
import os


@pytest.fixture
def empty_xvc_repo(monkeypatch, tmp_path):
    tmpdir = tmp_path / "xvc-py-test-repo"
    tmpdir.mkdir()
    monkeypatch.chdir(tmpdir)

    os.system("git init")
    assert ".git" in os.listdir()

    # xvc = Xvc(verbosity=4)
    xvc = Xvc()
    xvc.init()
    print(f"empty_xvc_repo: {xvc}")

    return xvc


@pytest.fixture
def xvc_repo_with_dir(empty_xvc_repo):
    assert ".git" in os.listdir()
    assert ".xvc" in os.listdir()

    os.system(
        "xvc-test-helper create-directory-tree --directories 3 --files 3 --seed 42"
    )

    print(f"empty_xvc_repo: {empty_xvc_repo}")
    return empty_xvc_repo


@pytest.fixture
def xvc_pipeline_single_step(xvc_repo_with_dir):
    assert ".git" in os.listdir()
    assert ".xvc" in os.listdir()

    xvc_repo_with_dir.pipeline().step().new(
        step_name="hello", command="echo 'hello xvc'"
    )

    print(f"xvc_repo_with_dir: {xvc_repo_with_dir}")

    return xvc_repo_with_dir


@pytest.fixture
def xvc_repo_with_people_csv(empty_xvc_repo):
    assert ".git" in os.listdir()
    assert ".xvc" in os.listdir()
    filename = "people.csv"

    with open(filename, "w") as f:
        f.write("""
Name,Sex,Age,Height,Weight
Alex,M,41,74,170
Bert,M,42,68,166
Carl,M,32,70,155
Dave,M,39,72,167
Elly,F,30,66,124
Fran,F,33,66,115
Gwen,F,26,64,121
Hank,M,30,71,158
Ivan,M,53,72,175
Jake,M,32,69,143
Kate,F,47,69,139
Luke,M,34,72,163
Myra,F,23,62,98
Neil,M,36,75,160
Omar,M,38,70,145
Page,F,31,67,135
Quin,M,29,71,176
Ruth,F,28,65,131
""")

    return empty_xvc_repo



import tempfile
import os


# How to set virtual env in lazyvim python
def test_storage_list(xvc_repo_with_dir):
    temp_dir = tempfile.mkdtemp()

    assert os.path.isdir(temp_dir)

    xvc_repo_with_dir.storage().new_local(name="local", path=temp_dir)

    assert "local" in str(xvc_repo_with_dir.storage().list())
    assert temp_dir in str(xvc_repo_with_dir.storage().list())


def test_storage_remove(xvc_repo_with_dir):
    # Create a temporary directory to be used with local storage
    temp_dir = tempfile.mkdtemp()

    assert os.path.isdir(temp_dir)

    xvc_repo_with_dir.storage().new_local(name="my-local-storage", path=temp_dir)

    assert "local" in str(xvc_repo_with_dir.storage().list())
    assert temp_dir in str(xvc_repo_with_dir.storage().list())

    xvc_repo_with_dir.storage().remove("my-local-storage")

    assert "local" not in str(xvc_repo_with_dir.storage().list())
    assert temp_dir not in str(xvc_repo_with_dir.storage().list())


#
#
#
# def test_storage_new(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_local(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_generic(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_rsync(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_s3(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_minio(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_digital_ocean(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_r2(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_gcs(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_wasabi(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_help(xvc_repo_with_dir):
#     assert False
#

# def test_file_send(xvc_repo_with_dir):
#     assert False
#
#
# def test_file_bring(xvc_repo_with_dir):
#     assert False
#
#
# def test_file_share(xvc_repo_with_dir):
#     assert False
#
# def test_file_remove_from_storage(xvc_repo_with_dir):
#     assert False
#



import os
import shutil


def test_file_hash(xvc_repo_with_dir):
    print(xvc_repo_with_dir.root(absolute=True))
    hash_string = xvc_repo_with_dir.file().hash("dir-0001/file-0002.bin")
    assert hash_string.startswith(
        "6432c99dec9e4a6c208ab78cfb58749ece5090fa8e279e6fd5a8cfd431e053f5"
    )


def test_file_track_symlink(xvc_repo_with_dir):
    assert ".git" in os.listdir()
    assert ".xvc" in os.listdir()

    xvc_repo_with_dir.file().track("dir-0001/file-0001.bin", recheck_method="symlink")
    assert os.path.islink("dir-0001/file-0001.bin")


def test_file_track_copy(xvc_repo_with_dir):
    print(os.listdir())
    xvc_repo_with_dir.file().track("dir-0001/file-0002.bin")
    assert os.path.isfile("dir-0001/file-0001.bin")
    assert os.path.isfile(
        ".xvc/b3/643/2c9/9dec9e4a6c208ab78cfb58749ece5090fa8e279e6fd5a8cfd431e053f5/0.bin"
    )


def test_file_recheck(xvc_repo_with_dir):
    xvc_repo_with_dir.file().track("dir-0001/file-0001.bin", recheck_method="symlink")
    os.remove("dir-0001/file-0001.bin")
    xvc_repo_with_dir.file().recheck("dir-0001/file-0001.bin")
    assert os.path.islink("dir-0001/file-0001.bin")


def test_file_carry_in(xvc_repo_with_dir):
    xvc_repo_with_dir.file().track("dir-0001/file-0001.bin")
    os.remove("dir-0001/file-0001.bin")
    shutil.copy("dir-0001/file-0002.bin", "dir-0001/file-0001.bin")
    xvc_repo_with_dir.file().carry_in("dir-0001/file-0001.bin")
    assert os.path.isfile(
        ".xvc/b3/643/2c9/9dec9e4a6c208ab78cfb58749ece5090fa8e279e6fd5a8cfd431e053f5/0.bin"
    )


def test_file_copy(xvc_repo_with_dir):
    xvc_repo_with_dir.file().track("dir-0001/file-0001.bin")
    os.remove("dir-0001/file-0001.bin")
    xvc_repo_with_dir.file().copy("dir-0001/file-0001.bin", "dir-0001/file-0005.bin")
    assert os.path.isfile("dir-0001/file-0005.bin")


def test_file_move(xvc_repo_with_dir):
    xvc_repo_with_dir.file().track("dir-0001/file-0002.bin", recheck_method="symlink")
    xvc_repo_with_dir.file().mv("dir-0001/file-0002.bin", "dir-0001/file-0005.bin")
    assert os.path.islink("dir-0001/file-0005.bin")
    assert not os.path.islink("dir-0001/file-0002.bin")


def test_file_list(xvc_repo_with_dir):
    file_list = xvc_repo_with_dir.file().list().split("\n")
    assert len([line for line in file_list if line.startswith("FX")]) == 9

    xvc_repo_with_dir.file().track("dir-0001/")
    file_list = xvc_repo_with_dir.file().list().split("\n")
    assert len([line for line in file_list if line.startswith("FC")]) == 3
    assert len([line for line in file_list if line.startswith("FX")]) == 6

    file_list = xvc_repo_with_dir.file().list(show_directories=True).split("\n")
    assert len([line for line in file_list if line.startswith("DX")]) == 3


def test_file_remove(xvc_repo_with_dir):
    xvc_repo_with_dir.file().track("dir-0001/")
    assert len(os.listdir(".xvc/b3/")) == 3

    xvc_repo_with_dir.file().remove("dir-0001/file-0001.bin", from_cache=True)
    assert len(os.listdir(".xvc/b3/")) == 2


def test_file_untrack(xvc_repo_with_dir):
    xvc_repo_with_dir.file().track("dir-0001/")
    assert len(os.listdir(".xvc/b3/")) == 3

    file_list = xvc_repo_with_dir.file().list().split("\n")
    assert len([line for line in file_list if line.startswith("FX")]) == 6
    assert len([line for line in file_list if line.startswith("FC")]) == 3

    xvc_repo_with_dir.file().untrack("dir-0001/file-0001.bin")

    file_list = xvc_repo_with_dir.file().list().split("\n")
    assert len([line for line in file_list if line.startswith("FX")]) == 7
    assert len([line for line in file_list if line.startswith("FC")]) == 2

    assert len(os.listdir(".xvc/b3/")) == 2



import os
import pytest
import yaml
import sqlite3
import time


def test_pipeline_list(empty_xvc_repo):
    pipeline_table = empty_xvc_repo.pipeline().list()
    expected = """
+---------+---------+
| Name    | Run Dir |
+===================+
| default |         |
+---------+---------+
""".strip()
    assert pipeline_table == expected


def test_pipeline_new(empty_xvc_repo):
    empty_xvc_repo.pipeline().new(pipeline_name="test", workdir="dir-0001")
    pipeline_table = empty_xvc_repo.pipeline().list()
    print(pipeline_table)
    expected = """
+---------+----------+
| Name    | Run Dir  |
+====================+
| default |          |
|---------+----------|
| test    | dir-0001 |
+---------+----------+
""".strip()
    assert pipeline_table == expected


def test_pipeline_update(empty_xvc_repo):
    empty_xvc_repo.pipeline().new(pipeline_name="test")
    empty_xvc_repo.pipeline().update(pipeline_name="test", workdir="dir-0001")
    pipeline_table = empty_xvc_repo.pipeline().list()
    print(pipeline_table)
    expected = """
+---------+----------+
| Name    | Run Dir  |
+====================+
| default |          |
|---------+----------|
| test    | dir-0001 |
+---------+----------+
""".strip()
    assert pipeline_table == expected


def test_pipeline_delete(empty_xvc_repo):
    empty_xvc_repo.pipeline().new(pipeline_name="test")
    empty_xvc_repo.pipeline().delete(pipeline_name="test")
    pipeline_table = empty_xvc_repo.pipeline().list()
    print(pipeline_table)
    expected = """
+---------+---------+
| Name    | Run Dir |
+===================+
| default |         |
+---------+---------+
""".strip()
    assert pipeline_table == expected


def test_pipeline_step_list(empty_xvc_repo):
    empty_xvc_repo.pipeline().step().new(step_name="hello", command="echo 'hello xvc'")
    pipeline_steps = empty_xvc_repo.pipeline().step().list(names_only=True)
    assert pipeline_steps.strip() == "hello"


def test_pipeline_step_new(empty_xvc_repo):
    empty_xvc_repo.pipeline().step().new(step_name="hello", command="echo 'hello xvc'")
    pipeline_steps = empty_xvc_repo.pipeline().step().list()
    assert pipeline_steps.strip() == "hello: echo 'hello xvc' (by_dependencies)"


def test_pipeline_step_update(empty_xvc_repo):
    empty_xvc_repo.pipeline().step().new(step_name="hello", command="echo 'hello xvc'")
    empty_xvc_repo.pipeline().step().update(
        step_name="hello", command="echo 'hello world'", when="always"
    )
    pipeline_steps = empty_xvc_repo.pipeline().step().list()
    assert pipeline_steps.strip() == "hello: echo 'hello world' (always)"


def test_pipeline_step_dependency_file(xvc_pipeline_single_step):
    pipeline = xvc_pipeline_single_step.pipeline()
    dependency_file = "dir-0001/file-0001.bin"
    pipeline.step().dependency(step_name="hello", file=dependency_file)
    first_run = pipeline.run()
    second_run = pipeline.run()
    os.system(f"xvc-test-helper generate-random-file {dependency_file}")
    third_run = pipeline.run()

    assert first_run == third_run
    assert second_run.strip() == ""


def test_pipeline_step_dependency_url(xvc_pipeline_single_step):
    pipeline = xvc_pipeline_single_step.pipeline()
    expected = """
[OUT] [hello] hello xvc
[DONE] [hello] (echo 'hello xvc')
    """.strip()

    pipeline.step().dependency(step_name="hello", url="https://xvc.dev")
    first_run = pipeline.run()
    second_run = pipeline.run()
    print(first_run.strip())
    assert first_run.strip() == expected
    assert second_run.strip() == ""


def test_pipeline_step_dependency_glob(xvc_pipeline_single_step):
    pipeline = xvc_pipeline_single_step.pipeline()
    dependency_file = "dir-0001/new-file.bin"
    pipeline.step().dependency(step_name="hello", glob="dir-0001/*.bin")

    first_run = pipeline.run()
    second_run = pipeline.run()
    os.system(f"xvc-test-helper generate-random-file {dependency_file}")
    third_run = pipeline.run()

    assert first_run == third_run
    assert second_run.strip() == ""


def test_pipeline_step_dependency_glob_items(xvc_repo_with_dir):
    pipeline = xvc_repo_with_dir.pipeline()
    dependency_dir = "dir-0001"
    dependency_file = f"{dependency_dir}/new-file.bin"
    xvc_repo_with_dir.pipeline().step().new(
        step_name="files", command='echo "ADDED_FILES: ${XVC_ADDED_GLOB_ITEMS}"'
    )
    pipeline.step().dependency(step_name="files", glob_items="dir-0001/*.bin")

    first_run = pipeline.run()
    second_run = pipeline.run()
    os.system(f"xvc-test-helper generate-random-file {dependency_file}")
    third_run = pipeline.run()

    print("FIRST RUN")
    print(first_run)
    print("SECOND RUN")
    print(second_run)
    print("THIRD RUN")
    print(third_run)

    assert (
        first_run.strip()
        == """
[OUT] [files] ADDED_FILES: dir-0001/file-0001.bin
dir-0001/file-0002.bin
dir-0001/file-0003.bin
[DONE] [files] (echo "ADDED_FILES: ${XVC_ADDED_GLOB_ITEMS}")
""".strip()
    )

    assert second_run.strip() == ""

    assert (
        third_run.strip()
        == """
[OUT] [files] ADDED_FILES: dir-0001/new-file.bin
[DONE] [files] (echo "ADDED_FILES: ${XVC_ADDED_GLOB_ITEMS}")
""".strip()
    )


def test_pipeline_step_dependency_step(xvc_pipeline_single_step):
    xvc_pipeline_single_step.pipeline().step().new(
        step_name="world", command="echo 'and the world'"
    )
    pipeline = xvc_pipeline_single_step.pipeline()
    pipeline.step().dependency(step_name="world", step="hello")
    first_run = pipeline.run()

    print(first_run)

    assert (
        first_run.strip()
        == """
[OUT] [hello] hello xvc
[DONE] [hello] (echo 'hello xvc')
[OUT] [world] and the world
[DONE] [world] (echo 'and the world')
        """.strip()
    )


def test_pipeline_step_dependency_regex(xvc_repo_with_people_csv):
    pipeline = xvc_repo_with_people_csv.pipeline()
    pipeline.step().new(step_name="a", command='echo "New names starting with A added"')
    pipeline.step().dependency(step_name="a", regex="people.csv:/^A.*$")

    first_run = pipeline.run()
    print(first_run)
    second_run = pipeline.run()
    print(second_run)
    with open("people.csv", "a") as f:
        f.write("Ali,M,13,74,170\n")
    third_run = pipeline.run()
    print(third_run)

    assert first_run.strip() == third_run.strip()
    assert second_run.strip() == ""


def test_pipeline_step_dependency_regex_items(xvc_repo_with_people_csv):
    pipeline = xvc_repo_with_people_csv.pipeline()
    pipeline.step().new(
        step_name="a", command='echo "Lines with A: ${XVC_ADDED_REGEX_ITEMS}"'
    )
    pipeline.step().dependency(step_name="a", regex_items="people.csv:/^A.*$")
    first_run = pipeline.run()
    print(first_run)
    second_run = pipeline.run()
    print(second_run)
    with open("people.csv", "a") as f:
        f.write("Ali,M,13,74,170\n")

    third_run = pipeline.run()
    print(third_run)

    assert second_run.strip() == ""

    assert (
        third_run.strip()
        == """
[OUT] [a] Lines with A: Ali,M,13,74,170
[DONE] [a] (echo "Lines with A: ${XVC_ADDED_REGEX_ITEMS}")
""".strip()
    )


def test_pipeline_step_dependency_line(xvc_repo_with_people_csv):
    pipeline = xvc_repo_with_people_csv.pipeline()
    pipeline.step().new(step_name="a", command='echo "New lines added to the file"')
    pipeline.step().dependency(step_name="a", lines="people.csv::10-")

    first_run = pipeline.run()
    print(first_run)
    second_run = pipeline.run()
    print(second_run)
    with open("people.csv", "a") as f:
        f.write("Ali,M,13,74,170\n")
    third_run = pipeline.run()
    print(third_run)

    assert first_run.strip() == third_run.strip()
    assert second_run.strip() == ""


def test_pipeline_step_dependency_line_items(xvc_repo_with_people_csv):
    pipeline = xvc_repo_with_people_csv.pipeline()
    pipeline.step().new(
        step_name="a", command='echo "Added lines: ${XVC_ADDED_LINE_ITEMS}"'
    )
    pipeline.step().dependency(step_name="a", line_items="people.csv::10-")
    first_run = pipeline.run()
    print(first_run)
    second_run = pipeline.run()
    print(second_run)
    with open("people.csv", "a") as f:
        f.write("Ali,M,13,74,170\n")

    third_run = pipeline.run()
    print(third_run)

    assert second_run.strip() == ""

    assert (
        third_run.strip()
        == """
[OUT] [a] Added lines: Ali,M,13,74,170
[DONE] [a] (echo "Added lines: ${XVC_ADDED_LINE_ITEMS}")
""".strip()
    )


def test_pipeline_step_dependency_param(empty_xvc_repo):
    filename = "params.yaml"

    with open(filename, "w") as f:
        f.write("""
param: value
database:
  server: example.com
  port: 5432
  connection:
    timeout: 5000
numeric_param: 13
""")
    pipeline = empty_xvc_repo.pipeline()
    pipeline.step().new(
        step_name="read-database-config", command=f"rg timeout {filename}"
    )
    pipeline.step().dependency(
        step_name="read-database-config",
        param=f"{filename}::database.connection.timeout",
    )
    first_run = pipeline.run()
    print(first_run)
    second_run = pipeline.run()
    print(second_run)

    update_yaml(filename, "database.connection.timeout", 10000)
    third_run = pipeline.run()
    assert (
        first_run.strip()
        == """
[OUT] [read-database-config]     timeout: 5000
[DONE] [read-database-config] (rg timeout params.yaml)
""".strip()
    )

    assert second_run.strip() == ""
    assert (
        third_run.strip()
        == """
[OUT] [read-database-config]     timeout: 10000
[DONE] [read-database-config] (rg timeout params.yaml)
""".strip()
    )


def update_yaml(file_path, key, new_value):
    # Read the existing YAML file
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)

    # Update the value
    keys = key.split(".")
    d = data
    for k in keys[:-1]:
        d = d.setdefault(k, {})
    d[keys[-1]] = new_value

    # Write the updated data back to the YAML file
    with open(file_path, "w") as file:
        yaml.safe_dump(data, file)


def test_pipeline_step_dependency_sqlite_query(empty_xvc_repo):
    filename = "people.db"
    db = sqlite3.connect(filename)

    db.execute("""
CREATE TABLE people (name, age, sex);
""")

    db.execute("""
INSERT INTO people VALUES ('Alice', 25, 'F'),
    ('Bob', 30, 'M'),
    ('Charlie', 35, 'M');
""")
    db.commit()

    pipeline = empty_xvc_repo.pipeline()
    pipeline.step().new(
        step_name="query", command=f'sqlite3 {filename} "SELECT AVG(age) FROM people;"'
    )
    pipeline.step().dependency(
        step_name="query",
        sqlite_file=filename,
        sqlite_query="SELECT COUNT(*) FROM people;",
    )
    first_run = pipeline.run()
    print(first_run)
    assert (
        first_run.strip()
        == """
[OUT] [query] 30.0
[DONE] [query] (sqlite3 people.db "SELECT AVG(age) FROM people;")
        """.strip()
    )
    second_run = pipeline.run()
    print(second_run)
    assert second_run.strip() == ""
    # NOTE: We are not changing the average age
    db.execute("INSERT INTO people VALUES ('David', 30, 'M');")
    db.commit()
    third_run = pipeline.run()
    print(third_run)
    assert first_run.strip() == third_run.strip()


# TODO: def test_pipeline_step_dependency_generic(xvc_repo_with_dir):
#   assert False
#
#
# TODO: def test_pipeline_dag(xvc_repo_with_dir):
#     assert False
#
#
# TODO: def test_pipeline_export(xvc_repo_with_dir):
#     assert False
#
#
# TODO: def test_pipeline_import(xvc_repo_with_dir):
#     assert False
#
